//============================================================================
// Name        : MGNoise.cpp
// Author      : Michal Gamrat
// Version     :
// Copyright   :
// Description : Autodesk Maya 2012 Plug-in deformer noise node
//============================================================================

#define LINUX

#define MAKE_INPUT(attr)        \
    CHECK_MSTATUS(attr.setKeyable(true) );              \
    CHECK_MSTATUS( attr.setStorable(true) );    \
    CHECK_MSTATUS( attr.setReadable(true) );    \
    CHECK_MSTATUS( attr.setWritable(true) );

#define MAKE_OUTPUT(attr)                       \
    CHECK_MSTATUS( attr.setKeyable(false) );    \
    CHECK_MSTATUS( attr.setStorable(false) );   \
    CHECK_MSTATUS( attr.setReadable(true) );    \
    CHECK_MSTATUS( attr.setWritable(false) );

#define MIN_OCT 1
#define MAX_OCT 10
#define MIN_PTS 5
#define MAX_PTS 500
#define MIN_SEED 0

#include <string.h>
#include <maya/MIOStream.h>
#include <math.h>

#include <maya/MPxDeformerNode.h>
#include <maya/MItGeometry.h>

#include <maya/MFnNumericAttribute.h>
#include <maya/MFnMatrixAttribute.h>
#include <maya/MFnMatrixData.h>

#include <maya/MFnPlugin.h>
#include <maya/MFnDependencyNode.h>

#include <maya/MTypeId.h>
#include <maya/MPlug.h>

#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MArrayDataHandle.h>

#include <maya/MPoint.h>
#include <maya/MVector.h>
#include <maya/MMatrix.h>

#include <maya/MDagModifier.h>
#include <maya/MFnTypedAttribute.h>
#include <maya/MFnNurbsCurve.h>
#include <maya/MPointArray.h>
#include <maya/MDoubleArray.h>
#include <maya/MDagPath.h>
#include <maya/MPlugArray.h>

#include <stdlib.h>

#include <maya/MPxLocatorNode.h>
#include <maya/MString.h>
#include <maya/MDagPath.h>
#include <maya/MTypeId.h>
#include <maya/MPlug.h>
#include <maya/MVector.h>
#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>
#include <maya/MColor.h>
#include <maya/M3dView.h>
#include <maya/MFnPlugin.h>
#include <maya/MDistance.h>
#include <maya/MFloatMatrix.h>
#include <maya/MFnUnitAttribute.h>
#include <maya/MFnPointArrayData.h>
#include <maya/MFnVectorArrayData.h>
#include <maya/MFnDoubleArrayData.h>
#include <maya/MVectorArray.h>
#include <maya/MFnVectorArrayData.h>
#include <maya/MGlobal.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "Test.cpp"
#include "perlin.h"

class noiseLocator : public MPxLocatorNode
{
	public:
								noiseLocator();
		virtual					~noiseLocator();
		virtual void            draw( M3dView & view, const MDagPath & path,
									  M3dView::DisplayStyle style,
									  M3dView::DisplayStatus status );

		virtual bool            isBounded() const;
		virtual MBoundingBox    boundingBox() const;

		virtual MStatus 		compute(const MPlug& plug, MDataBlock& dataBlock);

		static  void 			*creator();
		static  MStatus         initialize();

	public:
		static  MTypeId     id;
		static  MString     drawRegistrantId;

		static	MObject		oxCurvePoints;
		static	MObject		oxCrvLenIn;
		static	MObject		oxCrvLenOut;
		static	MObject		oxCrvPtsCountIn;
		static	MObject		oxCrvPtsCountOut;

		static	MObject		oyCurvePoints;
		static	MObject		oyCrvLenIn;
		static	MObject		oyCrvLenOut;
		static	MObject		oyCrvPtsCountIn;
		static	MObject		oyCrvPtsCountOut;

		static	MObject		ozCurvePoints;
		static	MObject		ozCrvLenIn;
		static	MObject		ozCrvLenOut;
		static	MObject		ozCrvPtsCountIn;
		static	MObject		ozCrvPtsCountOut;
	private:
		MVectorArray		getPoints(int) const;
};

MTypeId noiseLocator::id( 0x80007 );
MString noiseLocator::drawRegistrantId("NoiseLocatorNodePlugin");

MObject noiseLocator::oxCurvePoints;
MObject noiseLocator::oxCrvLenIn;
MObject noiseLocator::oxCrvLenOut;
MObject noiseLocator::oxCrvPtsCountIn;
MObject noiseLocator::oxCrvPtsCountOut;

MObject noiseLocator::oyCurvePoints;
MObject noiseLocator::oyCrvLenIn;
MObject noiseLocator::oyCrvLenOut;
MObject noiseLocator::oyCrvPtsCountIn;
MObject noiseLocator::oyCrvPtsCountOut;

MObject noiseLocator::ozCurvePoints;
MObject noiseLocator::ozCrvLenIn;
MObject noiseLocator::ozCrvLenOut;
MObject noiseLocator::ozCrvPtsCountIn;
MObject noiseLocator::ozCrvPtsCountOut;

noiseLocator::noiseLocator() {}
noiseLocator::~noiseLocator() {}

MStatus noiseLocator::initialize()
{
	MStatus stat;

	MFnTypedAttribute typedAttrFn;
	MVectorArray defaultVectArray;
	MFnVectorArrayData vectArrayDataFn;
	vectArrayDataFn.create( defaultVectArray );
	typedAttrFn.create( "oxCurvePoints", "oxP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
	oxCurvePoints = typedAttrFn.object();
	MAKE_INPUT(typedAttrFn);
	addAttribute( oxCurvePoints );

	MFnNumericAttribute numAttr;

	oxCrvLenIn = numAttr.create("oxCrvLenIn", "xli", MFnNumericData::kFloat);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(1.0);
	addAttribute(oxCrvLenIn);
	oxCrvLenOut = numAttr.create("oxCrvLenOut", "xlo", MFnNumericData::kFloat);
	MAKE_OUTPUT(numAttr);
	stat = addAttribute(oxCrvLenOut);
	CHECK_MSTATUS(stat);
	stat = attributeAffects(oxCrvLenIn, oxCrvLenOut);
	CHECK_MSTATUS(stat);

	oxCrvPtsCountIn = numAttr.create("oxCrvPtsCountIn", "xci", MFnNumericData::kInt);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(30);
	numAttr.setMin(MIN_PTS);
	numAttr.setMax(MAX_PTS);
	addAttribute(oxCrvPtsCountIn);
	oxCrvPtsCountOut = numAttr.create("oxCrvPtsCountOut", "xco", MFnNumericData::kInt);
	MAKE_OUTPUT(numAttr);
	addAttribute(oxCrvPtsCountOut);
	attributeAffects(oxCrvPtsCountIn, oxCrvPtsCountOut);

	//oy

	vectArrayDataFn.create( defaultVectArray );
	typedAttrFn.create( "oyCurvePoints", "oyP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
	oyCurvePoints = typedAttrFn.object();
	MAKE_INPUT(typedAttrFn);
	addAttribute( oyCurvePoints );

	oyCrvLenIn = numAttr.create("oyCrvLenIn", "yli", MFnNumericData::kFloat);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(1.0);
	addAttribute(oyCrvLenIn);
	oyCrvLenOut = numAttr.create("oyCrvLenOut", "ylo", MFnNumericData::kFloat);
	MAKE_OUTPUT(numAttr);
	stat = addAttribute(oyCrvLenOut);
	CHECK_MSTATUS(stat);
	stat = attributeAffects(oyCrvLenIn, oyCrvLenOut);
	CHECK_MSTATUS(stat);

	oyCrvPtsCountIn = numAttr.create("oyCrvPtsCountIn", "yci", MFnNumericData::kInt);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(30);
	numAttr.setMin(MIN_PTS);
	numAttr.setMax(MAX_PTS);
	addAttribute(oyCrvPtsCountIn);
	oyCrvPtsCountOut = numAttr.create("oyCrvPtsCountOut", "yco", MFnNumericData::kInt);
	MAKE_OUTPUT(numAttr);
	addAttribute(oyCrvPtsCountOut);
	attributeAffects(oyCrvPtsCountIn, oyCrvPtsCountOut);

	//oz

	vectArrayDataFn.create( defaultVectArray );
	typedAttrFn.create( "ozCurvePoints", "ozP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
	ozCurvePoints = typedAttrFn.object();
	MAKE_INPUT(typedAttrFn);
	addAttribute( ozCurvePoints );

	ozCrvLenIn = numAttr.create("ozCrvLenIn", "zli", MFnNumericData::kFloat);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(1.0);
	addAttribute(ozCrvLenIn);
	ozCrvLenOut = numAttr.create("ozCrvLenOut", "zlo", MFnNumericData::kFloat);
	MAKE_OUTPUT(numAttr);
	stat = addAttribute(ozCrvLenOut);
	CHECK_MSTATUS(stat);
	stat = attributeAffects(ozCrvLenIn, ozCrvLenOut);
	CHECK_MSTATUS(stat);

	ozCrvPtsCountIn = numAttr.create("ozCrvPtsCountIn", "zci", MFnNumericData::kInt);
	MAKE_INPUT(numAttr);
	numAttr.setDefault(30);
	numAttr.setMin(MIN_PTS);
	numAttr.setMax(MAX_PTS);
	addAttribute(ozCrvPtsCountIn);
	ozCrvPtsCountOut = numAttr.create("ozCrvPtsCountOut", "zco", MFnNumericData::kInt);
	MAKE_OUTPUT(numAttr);
	addAttribute(ozCrvPtsCountOut);
	attributeAffects(ozCrvPtsCountIn, ozCrvPtsCountOut);

	return MStatus::kSuccess;
}

MStatus noiseLocator::compute(const MPlug& plug, MDataBlock& block)
{
	MDataHandle handle;
	bool plugFound = true;

	if(plug == oxCrvLenOut)
	{
		handle = block.inputValue(oxCrvLenIn);
		float oxCrvLen = handle.asFloat();
		handle = block.outputValue(oxCrvLenOut);
		handle.set(oxCrvLen);
	}else if(plug == oxCrvPtsCountOut){
		handle = block.inputValue(oxCrvPtsCountIn);
		int count = handle.asInt();
		handle = block.outputValue(oxCrvPtsCountOut);
		handle.set(count);
	}else if(plug == oyCrvLenOut){
		handle = block.inputValue(oyCrvLenIn);
		float oyCrvLen = handle.asFloat();
		handle = block.outputValue(oyCrvLenOut);
		handle.set(oyCrvLen);
	}else if(plug == oyCrvPtsCountOut){
		handle = block.inputValue(oyCrvPtsCountIn);
		int count = handle.asInt();
		handle = block.outputValue(oyCrvPtsCountOut);
		handle.set(count);
	}else if(plug == ozCrvLenOut){
		handle = block.inputValue(ozCrvLenIn);
		float ozCrvLen = handle.asFloat();
		handle = block.outputValue(ozCrvLenOut);
		handle.set(ozCrvLen);
	}else if(plug == ozCrvPtsCountOut){
		handle = block.inputValue(ozCrvPtsCountIn);
		int count = handle.asInt();
		handle = block.outputValue(ozCrvPtsCountOut);
		handle.set(count);
	}else{
		plugFound = false;
	}

	if(plugFound)
	{
		block.setClean(plug);
		return MStatus::kSuccess;
	}

	return MStatus::kUnknownParameter;

}

MVectorArray noiseLocator::getPoints(int axis) const
{
	MStatus stat;
	MObject thisNode = thisMObject();
	MFnDagNode dagFn(thisNode);
	MPlug axisPlug;

	if(axis == 0)
	{
		axisPlug = dagFn.findPlug(oxCurvePoints, &stat);
	}else if(axis == 1)
	{
		axisPlug = dagFn.findPlug(oyCurvePoints, &stat);
	}else{
		axisPlug = dagFn.findPlug(ozCurvePoints, &stat);
	}

	if (!stat) {
		stat.perror("plug");
	}

	MObject vectorArrayData;
	stat = axisPlug.getValue(vectorArrayData);

	if (!stat) {
		stat.perror("getValue");
	}

	MFnVectorArrayData fnVAD(vectorArrayData, &stat);
	if (!stat) {
		stat.perror("MFnVectorArrayData");
	}

	MVectorArray finalArray = fnVAD.array();
	return finalArray;
}


void noiseLocator::draw( M3dView & view, const MDagPath & /*path*/, M3dView::DisplayStyle style, M3dView::DisplayStatus status )
{

    MObject thisNode = thisMObject();
    view.beginGL();
    glPushAttrib(GL_CURRENT_BIT);
    glBegin( GL_LINES );

    glColor3f(1.0f, 0.0f, 0.0f);

    glVertex3f(-1,0,0);
    glVertex3f(1,0,0);
    glVertex3f(0,0,-1);
    glVertex3f(0,0,1);
    glVertex3f(0,-1,0);
    glVertex3f(0,1,0);

    glEnd();

    glColor3f(0.0f, 0.0f, 1.0f);

    glBegin(GL_LINE_STRIP);

    int a;
    for(a=0; a<3; a++)
    {
    	MVectorArray pts = noiseLocator::getPoints(a);
		int i;
		int last = pts.length();
		for ( i = 0; i < last; i++ ) {
			glVertex3f( pts[i].x, pts[i].y,  pts[i].z);
		}

    }

    glEnd();
    glPopAttrib();
    view.endGL();
}

bool noiseLocator::isBounded() const
{
    return true;
}

MBoundingBox noiseLocator::boundingBox() const
{
    // Get the size
    //
    MObject thisNode = thisMObject();

    double multiplier = 5;

    MPoint corner1( -1, 0.0, -1 );
    MPoint corner2( 1, 0.0, 1 );

    corner1 = corner1 * multiplier;
    corner2 = corner2 * multiplier;

    return MBoundingBox( corner1, corner2 );
}

void* noiseLocator::creator()
{
    return new noiseLocator();
}

class mgNoise : public MPxDeformerNode
{
	public:
		mgNoise();
		virtual ~mgNoise();

		static  void* creator();
		static  MStatus initialize();

		virtual MStatus deform(MDataBlock& block, MItGeometry& iter, const MMatrix& mat, unsigned int multiIndex);
		virtual MObject& accessoryAttribute() const;
		//virtual MStatus  setDependentsDirty( const MPlug& plugBeingDirtied, MPlugArray &affectedPlugs );
		virtual MStatus compute(const MPlug& plug, MDataBlock& dataBlock);


		virtual MStatus accessoryNodeSetup(MDagModifier& cmd);

	public:
		static  MTypeId     id;
		static	MObject		deformSpace;

		static	MObject		oxVisualisationPoints;
		static	MObject		oxLen;
		static	MObject		oxPtsCount;
		static	MObject		oxNoiseFreq;
		static	MObject		oxNoiseAmp;
		static	MObject		oxNoiseOctaves;
		static	MObject		oxNoiseSeed;

		static	MObject		oyVisualisationPoints;
		static	MObject		oyLen;
		static	MObject		oyPtsCount;
		static	MObject		oyNoiseFreq;
		static	MObject		oyNoiseAmp;
		static	MObject		oyNoiseOctaves;
		static	MObject		oyNoiseSeed;

		static	MObject		ozVisualisationPoints;
		static	MObject		ozLen;
		static	MObject		ozPtsCount;
		static	MObject		ozNoiseFreq;
		static	MObject		ozNoiseAmp;
		static	MObject		ozNoiseOctaves;
		static	MObject		ozNoiseSeed;


};

MTypeId     mgNoise::id( 0x000A1 );
MObject     mgNoise::deformSpace;

MObject     mgNoise::oxVisualisationPoints;
MObject     mgNoise::oxLen;
MObject     mgNoise::oxPtsCount;

MObject     mgNoise::oxNoiseFreq;
MObject     mgNoise::oxNoiseAmp;
MObject     mgNoise::oxNoiseOctaves;
MObject     mgNoise::oxNoiseSeed;


MObject     mgNoise::oyVisualisationPoints;
MObject     mgNoise::oyLen;
MObject     mgNoise::oyPtsCount;

MObject     mgNoise::oyNoiseFreq;
MObject     mgNoise::oyNoiseAmp;
MObject     mgNoise::oyNoiseOctaves;
MObject     mgNoise::oyNoiseSeed;


MObject     mgNoise::ozVisualisationPoints;
MObject     mgNoise::ozLen;
MObject     mgNoise::ozPtsCount;

MObject     mgNoise::ozNoiseFreq;
MObject     mgNoise::ozNoiseAmp;
MObject     mgNoise::ozNoiseOctaves;
MObject     mgNoise::ozNoiseSeed;



mgNoise::mgNoise() {}
mgNoise::~mgNoise() {}

void* mgNoise::creator()
{
        return new mgNoise();
}

MStatus mgNoise::initialize()
{
        MStatus stat;
        MFnMatrixAttribute mAttr;
        deformSpace = mAttr.create("deformSpace", "dSp");
        mAttr.setStorable(false);
        addAttribute(deformSpace);
        attributeAffects( mgNoise::deformSpace, mgNoise::outputGeom );

        MFnTypedAttribute typedAttrFn;
		MVectorArray defaultVectArray;
		MFnVectorArrayData vectArrayDataFn;
		vectArrayDataFn.create( defaultVectArray );
		typedAttrFn.create( "oxVisualisationPoints", "oxP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
		CHECK_MSTATUS(stat);

		typedAttrFn.setStorable(false);
		typedAttrFn.setConnectable(true);
		//typedAttrFn.setCached(false);

		oxVisualisationPoints = typedAttrFn.object();
		//MAKE_OUTPUT(typedAttrFn);
		stat = addAttribute( oxVisualisationPoints );
		CHECK_MSTATUS(stat);

		vectArrayDataFn.create( defaultVectArray );
		typedAttrFn.create( "oyVisualisationPoints", "oyP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
		CHECK_MSTATUS(stat);

		typedAttrFn.setStorable(false);
		typedAttrFn.setConnectable(true);
		oyVisualisationPoints = typedAttrFn.object();
		stat = addAttribute( oyVisualisationPoints );
		CHECK_MSTATUS(stat);

		vectArrayDataFn.create( defaultVectArray );
		typedAttrFn.create( "ozVisualisationPoints", "ozP", MFnData::kVectorArray, vectArrayDataFn.object(), &stat );
		CHECK_MSTATUS(stat);

		typedAttrFn.setStorable(false);
		typedAttrFn.setConnectable(true);
		ozVisualisationPoints = typedAttrFn.object();
		stat = addAttribute( ozVisualisationPoints );
		CHECK_MSTATUS(stat);

		MFnNumericAttribute nAttr;

		//ox

		oxLen = nAttr.create("oxLength", "oxL", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		addAttribute(oxLen);
		attributeAffects(mgNoise::oxLen, mgNoise::oxVisualisationPoints);

		oxPtsCount = nAttr.create("oxPointsCount", "xpc", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setMin(MIN_PTS);
		nAttr.setMax(MAX_PTS);
		addAttribute(oxPtsCount);
		attributeAffects(mgNoise::oxPtsCount, mgNoise::oxVisualisationPoints);

		oxNoiseOctaves = nAttr.create("oxNoiseOctaves", "xno", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(4);
		nAttr.setMin(MIN_OCT);
		nAttr.setMax(MAX_OCT);
		addAttribute(oxNoiseOctaves);
		attributeAffects(mgNoise::oxNoiseOctaves, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseOctaves, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseOctaves, mgNoise::outputGeom );

		oxNoiseSeed  = nAttr.create("oxNoiseSeed", "xns", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(0);
		nAttr.setMin(MIN_SEED);
		addAttribute(oxNoiseSeed);
		attributeAffects(mgNoise::oxNoiseSeed, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseSeed, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseSeed, mgNoise::outputGeom );

		oxNoiseFreq = nAttr.create("oxNoiseFreq", "xnf", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1.0);
		addAttribute(oxNoiseFreq);
		attributeAffects(mgNoise::oxNoiseFreq, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseFreq, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseFreq, mgNoise::outputGeom );

		oxNoiseAmp = nAttr.create("oxNoiseAmp", "xna", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(0.0);
		addAttribute(oxNoiseAmp);
		attributeAffects(mgNoise::oxNoiseAmp, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseAmp, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oxNoiseAmp, mgNoise::outputGeom );


		//oy

		oyLen = nAttr.create("oyLength", "oyL", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		addAttribute(oyLen);
		attributeAffects(mgNoise::oyLen, mgNoise::oyVisualisationPoints);

		oyPtsCount = nAttr.create("oyPointsCount", "ypc", MFnNumericData::kInt);
		nAttr.setMin(MIN_PTS);
		nAttr.setMax(MAX_PTS);
		MAKE_INPUT(nAttr);
		addAttribute(oyPtsCount);
		attributeAffects(mgNoise::oyPtsCount, mgNoise::oyVisualisationPoints);

		oyNoiseOctaves = nAttr.create("oyNoiseOctaves", "yno", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(4);
		nAttr.setMin(MIN_OCT);
		nAttr.setMax(MAX_OCT);
		addAttribute(oyNoiseOctaves);
		attributeAffects(mgNoise::oyNoiseOctaves, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseOctaves, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseOctaves, mgNoise::outputGeom );

		oyNoiseSeed  = nAttr.create("oyNoiseSeed", "yns", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1);
		nAttr.setMin(MIN_SEED);
		addAttribute(oyNoiseSeed);
		attributeAffects(mgNoise::oyNoiseSeed, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseSeed, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseSeed, mgNoise::outputGeom );

		oyNoiseFreq = nAttr.create("oyNoiseFreq", "ynf", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1.0);
		addAttribute(oyNoiseFreq);
		attributeAffects(mgNoise::oyNoiseFreq, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseFreq, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseFreq, mgNoise::outputGeom );

		oyNoiseAmp = nAttr.create("oyNoiseAmp", "yna", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(0.0);
		addAttribute(oyNoiseAmp);
		attributeAffects(mgNoise::oyNoiseAmp, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseAmp, mgNoise::ozVisualisationPoints);
		attributeAffects(mgNoise::oyNoiseAmp, mgNoise::outputGeom );

		//oz

		ozLen = nAttr.create("ozLength", "ozL", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		addAttribute(ozLen);
		attributeAffects(mgNoise::ozLen, mgNoise::ozVisualisationPoints);

		ozPtsCount = nAttr.create("ozPointsCount", "zpc", MFnNumericData::kInt);
		nAttr.setMin(MIN_PTS);
		nAttr.setMax(MAX_PTS);
		MAKE_INPUT(nAttr);
		addAttribute(ozPtsCount);
		attributeAffects(mgNoise::ozPtsCount, mgNoise::ozVisualisationPoints);

		ozNoiseOctaves = nAttr.create("ozNoiseOctaves", "zno", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(4);
		nAttr.setMin(MIN_OCT);
		nAttr.setMax(MAX_OCT);
		addAttribute(ozNoiseOctaves);
		attributeAffects(mgNoise::ozNoiseOctaves, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseOctaves, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseOctaves, mgNoise::outputGeom );

		ozNoiseSeed  = nAttr.create("ozNoiseSeed", "zns", MFnNumericData::kInt);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(2);
		nAttr.setMin(MIN_SEED);
		addAttribute(ozNoiseSeed);
		attributeAffects(mgNoise::ozNoiseSeed, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseSeed, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseSeed, mgNoise::outputGeom );

		ozNoiseFreq = nAttr.create("ozNoiseFreq", "znf", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1.0);
		addAttribute(ozNoiseFreq);
		attributeAffects(mgNoise::ozNoiseFreq, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseFreq, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseFreq, mgNoise::outputGeom );

		ozNoiseAmp = nAttr.create("ozNoiseAmp", "zna", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setDefault(0.0);
		addAttribute(ozNoiseAmp);
		attributeAffects(mgNoise::ozNoiseAmp, mgNoise::oxVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseAmp, mgNoise::oyVisualisationPoints);
		attributeAffects(mgNoise::ozNoiseAmp, mgNoise::outputGeom );


        return MStatus::kSuccess;
}

MStatus mgNoise::compute(const MPlug& plug, MDataBlock& block)
{

		bool plugFound = true;

		if(plug == oxVisualisationPoints)
		{
			MDataHandle handle = block.inputValue(oxLen);
			float length = handle.asFloat();
			handle = block.inputValue(oxPtsCount);
			int count = handle.asInt();

			if(count<MIN_PTS)
			{
				count = MIN_PTS;
			}else if(count>MAX_PTS){
				count = MAX_PTS;
			}

			handle = block.inputValue(oyNoiseFreq);
			float yNoiseF = handle.asFloat();

			handle = block.inputValue(oyNoiseOctaves);
			int yNoiseO = handle.asInt();

			if(yNoiseO<MIN_OCT)
			{
				yNoiseO = MIN_OCT;
			}else if(yNoiseO>MAX_OCT){
				yNoiseO = MAX_OCT;
			}

			handle = block.inputValue(oyNoiseSeed);
			int yNoiseS = handle.asInt();

			if(yNoiseS<MIN_SEED)
			{
				yNoiseS = MIN_SEED;
			}

			handle = block.inputValue(oyNoiseAmp);
			float yNoiseA = handle.asFloat();

			Perlin perlinNoiseY(yNoiseO, yNoiseF, yNoiseA, yNoiseS);

			//Z-Noise

			handle = block.inputValue(ozNoiseFreq);
			float zNoiseF = handle.asFloat();

			handle = block.inputValue(ozNoiseOctaves);
			int zNoiseO = handle.asInt();

			if(zNoiseO<MIN_OCT)
			{
				zNoiseO = MIN_OCT;
			}else if(zNoiseO>MAX_OCT){
				zNoiseO = MAX_OCT;
			}

			handle = block.inputValue(ozNoiseSeed);
			int zNoiseS = handle.asInt();

			if(zNoiseS<MIN_SEED)
			{
				zNoiseS = MIN_SEED;
			}

			handle = block.inputValue(ozNoiseAmp);
			float zNoiseA = handle.asFloat();

			Perlin perlinNoiseZ(zNoiseO, zNoiseF, zNoiseA, zNoiseS);

			MVectorArray va;

			int i;
			float dx = length/((float)count);


			for(i=0; i<count; i++)
			{
				float x = dx*i;
				float y = perlinNoiseY.Get(x,0);
				float z = perlinNoiseZ.Get(x, y);
				MVector v(x, y, z);
				va.append(v);
			}

			MFnVectorArrayData fnVectArr;
			MObject objArr = fnVectArr.create(va);

			MDataHandle oxPtsHandle = block.outputValue(oxVisualisationPoints);
			oxPtsHandle.set(objArr);
		}else if(plug == oyVisualisationPoints){
			MDataHandle handle = block.inputValue(oyLen);
			float length = handle.asFloat();
			handle = block.inputValue(oyPtsCount);
			int count = handle.asInt();

			if(count<MIN_PTS)
			{
				count = MIN_PTS;
			}else if(count>MAX_PTS){
				count = MAX_PTS;
			}

			handle = block.inputValue(oxNoiseFreq);
			float xNoiseF = handle.asFloat();

			handle = block.inputValue(oxNoiseOctaves);
			int xNoiseO = handle.asInt();

			if(xNoiseO<MIN_OCT)
			{
				xNoiseO = MIN_OCT;
			}else if(xNoiseO>MAX_OCT){
				xNoiseO = MAX_OCT;
			}

			handle = block.inputValue(oxNoiseSeed);
			int xNoiseS = handle.asInt();

			if(xNoiseS<MIN_SEED)
			{
				xNoiseS = MIN_SEED;
			}

			handle = block.inputValue(oxNoiseAmp);
			float xNoiseA = handle.asFloat();

			Perlin perlinNoiseX(xNoiseO, xNoiseF, xNoiseA, xNoiseS);

			//Z-Noise

			handle = block.inputValue(ozNoiseFreq);
			float zNoiseF = handle.asFloat();

			handle = block.inputValue(ozNoiseOctaves);
			int zNoiseO = handle.asInt();

			if(zNoiseO<MIN_OCT)
			{
				zNoiseO = MIN_OCT;
			}else if(zNoiseO>MAX_OCT){
				zNoiseO = MAX_OCT;
			}

			handle = block.inputValue(ozNoiseSeed);
			int zNoiseS = handle.asInt();

			if(zNoiseS<MIN_SEED)
			{
				zNoiseS = MIN_SEED;
			}

			handle = block.inputValue(ozNoiseAmp);
			float zNoiseA = handle.asFloat();

			Perlin perlinNoiseZ(zNoiseO, zNoiseF, zNoiseA, zNoiseS);

			MVectorArray va;

			int i;
			float dy = length/((float)count);


			for(i=0; i<count; i++)
			{
				float y = dy*i;
				float x = perlinNoiseX.Get(y,0);
				float z = perlinNoiseZ.Get(x, y);
				MVector v(x, y, z);
				va.append(v);
			}

			MFnVectorArrayData fnVectArr;
			MObject objArr = fnVectArr.create(va);
			MDataHandle oyPtsHandle = block.outputValue(oyVisualisationPoints);
			oyPtsHandle.set(objArr);
		}else if(plug == ozVisualisationPoints){
			MDataHandle handle = block.inputValue(ozLen);
			float length = handle.asFloat();
			handle = block.inputValue(ozPtsCount);
			int count = handle.asInt();

			if(count<MIN_PTS)
			{
				count = MIN_PTS;
			}else if(count>MAX_PTS){
				count = MAX_PTS;
			}

			handle = block.inputValue(oyNoiseFreq);
			float yNoiseF = handle.asFloat();

			handle = block.inputValue(oyNoiseOctaves);
			int yNoiseO = handle.asInt();

			if(yNoiseO<MIN_OCT)
			{
				yNoiseO = MIN_OCT;
			}else if(yNoiseO>MAX_OCT){
				yNoiseO = MAX_OCT;
			}

			handle = block.inputValue(oyNoiseSeed);
			int yNoiseS = handle.asInt();

			if(yNoiseS<MIN_SEED)
			{
				yNoiseS = MIN_SEED;
			}

			handle = block.inputValue(oyNoiseAmp);
			float yNoiseA = handle.asFloat();

			//X-Noise

			handle = block.inputValue(oxNoiseFreq);
			float xNoiseF = handle.asFloat();

			handle = block.inputValue(oxNoiseOctaves);
			int xNoiseO = handle.asInt();

			if(xNoiseO<MIN_OCT)
			{
				xNoiseO = MIN_OCT;
			}else if(xNoiseO>MAX_OCT){
				xNoiseO = MAX_OCT;
			}

			handle = block.inputValue(oxNoiseSeed);
			int xNoiseS = handle.asInt();

			if(xNoiseS<MIN_SEED)
			{
				xNoiseS = MIN_SEED;
			}

			handle = block.inputValue(oxNoiseAmp);
			float xNoiseA = handle.asFloat();

			Perlin perlinNoiseX(xNoiseO, xNoiseF, xNoiseA, xNoiseS);

			Perlin perlinNoiseY(yNoiseO, yNoiseF, yNoiseA, yNoiseS);

			MVectorArray va;

			int i;
			float dz = length/((float)count);

			for(i=0; i<count; i++)
			{
				float z = dz*i;
				float y = perlinNoiseY.Get(0, z);
				float x = perlinNoiseX.Get(y, z);
				MVector v(x, y, z);
				va.append(v);
			}

			MFnVectorArrayData fnVectArr;
			MObject objArr = fnVectArr.create(va);

			MDataHandle ozPtsHandle = block.outputValue(ozVisualisationPoints);
			ozPtsHandle.set(objArr);
		}else{
			plugFound = false;
		}

		if(plugFound)
		{
			block.setClean(plug);
			return MStatus::kSuccess;
		}

		return MStatus::kUnknownParameter;

}


MStatus mgNoise::deform( MDataBlock& block, MItGeometry& iter, const MMatrix& m, unsigned int multiIndex)
{
		MStatus status;
        MDataHandle matData = block.inputValue(deformSpace, &status);

        if (MS::kSuccess != status) return status;

        MMatrix mat = matData.asMatrix();
        MMatrix invMat = mat.inverse();

		matData = block.inputValue(oyNoiseOctaves, &status);
		CHECK_MSTATUS(status);
		int yNoiseO = matData.asInt();

		if(yNoiseO<MIN_OCT)
		{
			yNoiseO = MIN_OCT;
		}else if(yNoiseO>MAX_OCT){
			yNoiseO = MAX_OCT;
		}

		matData = block.inputValue(oyNoiseFreq, &status);
		CHECK_MSTATUS(status);
		float yNoiseF = matData.asFloat();

		matData = block.inputValue(oyNoiseAmp, &status);
		CHECK_MSTATUS(status);
		float yNoiseA = matData.asFloat();

		matData = block.inputValue(oyNoiseSeed, &status);
		CHECK_MSTATUS(status);
		int yNoiseS = matData.asInt();

		if(yNoiseS<MIN_SEED)
		{
			yNoiseS = MIN_SEED;
		}

		Perlin perlinNoiseY(yNoiseO, yNoiseF, yNoiseA, yNoiseS);

		//X-Noise

		matData = block.inputValue(oxNoiseOctaves, &status);
		CHECK_MSTATUS(status);
		int xNoiseO = matData.asInt();

		if(xNoiseO<MIN_OCT)
		{
			xNoiseO = MIN_OCT;
		}else if(xNoiseO>MAX_OCT){
			xNoiseO = MAX_OCT;
		}

		matData = block.inputValue(oxNoiseFreq, &status);
		CHECK_MSTATUS(status);
		float xNoiseF = matData.asFloat();

		matData = block.inputValue(oxNoiseAmp, &status);
		CHECK_MSTATUS(status);
		float xNoiseA = matData.asFloat();

		matData = block.inputValue(oxNoiseSeed, &status);
		CHECK_MSTATUS(status);
		int xNoiseS = matData.asInt();

		if(xNoiseS<MIN_SEED)
		{
			xNoiseS = MIN_SEED;
		}

		Perlin perlinNoiseX(xNoiseO, xNoiseF, xNoiseA, xNoiseS);

		//Z-Noise

		matData = block.inputValue(ozNoiseOctaves, &status);
		CHECK_MSTATUS(status);
		int zNoiseO = matData.asInt();

		if(zNoiseO<MIN_OCT)
		{
			zNoiseO = MIN_OCT;
		}else if(zNoiseO>MAX_OCT){
			zNoiseO = MAX_OCT;
		}

		matData = block.inputValue(ozNoiseFreq, &status);
		CHECK_MSTATUS(status);
		float zNoiseF = matData.asFloat();

		matData = block.inputValue(ozNoiseAmp, &status);
		CHECK_MSTATUS(status);
		float zNoiseA = matData.asFloat();

		matData = block.inputValue(ozNoiseSeed, &status);
		CHECK_MSTATUS(status);
		int zNoiseS = matData.asInt();

		if(zNoiseS<MIN_SEED)
		{
			zNoiseS = MIN_SEED;
		}

		Perlin perlinNoiseZ(zNoiseO, zNoiseF, zNoiseA, zNoiseS);

		MMatrix mi = m.inverse();

        for ( ; !iter.isDone(); iter.next()) {
			MPoint pt = iter.position();

			pt *= invMat;

			pt.y += perlinNoiseY.Get(pt.x, pt.z);
			pt.x += perlinNoiseX.Get(pt.y, pt.z);
			pt.z += perlinNoiseZ.Get(pt.x, pt.y);

			pt *= mat;

			iter.setPosition(pt);
        }

        return status;
}

/* override */ MObject& mgNoise::accessoryAttribute() const
{
		return deformSpace;
}

const double DEG_TO_RAD = M_PI /180.0;
const double RAD_TO_DEG = 180.0/ M_PI;

inline double degToRad(const double d)
{
	return d * DEG_TO_RAD;
}

inline double radToDeg(const double d)
{
	return d * RAD_TO_DEG;
}
/* override */ MStatus mgNoise::accessoryNodeSetup(MDagModifier& cmd)
{
	MStatus stat;
	MObject transformObj = cmd.createNode(MString("transform"),  MObject::kNullObj, &stat);
	MObject shapeObj = cmd.createNode(noiseLocator::id,  transformObj, &stat);
	MFnDependencyNode locFn( transformObj );
	MObject attr = locFn.attribute("matrix");
	stat = cmd.connect(transformObj, attr, thisMObject(), deformSpace);
	CHECK_MSTATUS(stat);
	locFn.setObject(shapeObj);

	//ox

	attr = locFn.attribute("oxCrvLenOut", &stat);
	CHECK_MSTATUS(stat);
	stat = cmd.connect(shapeObj, attr, thisMObject(), oxLen);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("oxCrvPtsCountOut");
	stat = cmd.connect(shapeObj, attr, thisMObject(), oxPtsCount);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("oxCurvePoints");
	stat = cmd.connect(thisMObject(), oxVisualisationPoints, shapeObj, attr);
	CHECK_MSTATUS(stat);

	//oy

	attr = locFn.attribute("oyCrvLenOut", &stat);
	CHECK_MSTATUS(stat);
	stat = cmd.connect(shapeObj, attr, thisMObject(), oyLen);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("oyCrvPtsCountOut");
	stat = cmd.connect(shapeObj, attr, thisMObject(), oyPtsCount);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("oyCurvePoints");
	stat = cmd.connect(thisMObject(), oyVisualisationPoints, shapeObj, attr);
	CHECK_MSTATUS(stat);

	//oz

	attr = locFn.attribute("ozCrvLenOut", &stat);
	CHECK_MSTATUS(stat);
	stat = cmd.connect(shapeObj, attr, thisMObject(), ozLen);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("ozCrvPtsCountOut");
	stat = cmd.connect(shapeObj, attr, thisMObject(), ozPtsCount);
	CHECK_MSTATUS(stat);

	attr = locFn.attribute("ozCurvePoints");
	stat = cmd.connect(thisMObject(), ozVisualisationPoints, shapeObj, attr);
	CHECK_MSTATUS(stat);

	return stat;
}

MStatus initializePlugin( MObject obj )
{
		MStatus result;
		MFnPlugin plugin( obj, PLUGIN_COMPANY, "3.0", "Any");

		result = plugin.registerNode("noiseLocator", noiseLocator::id, &noiseLocator::creator, &noiseLocator::initialize, MPxNode::kLocatorNode);
		if (!result) {
			result.perror("registerNode");
			return result;
		}

		result = plugin.registerNode( "mgNoise", mgNoise::id, mgNoise::creator, mgNoise::initialize, MPxNode::kDeformerNode );
		return result;
}

MStatus uninitializePlugin( MObject obj)
{
		MStatus result;
		MFnPlugin plugin( obj );

		result = plugin.deregisterNode( noiseLocator::id );
		if (!result) {
			result.perror("deregisterNode");
			return result;
		}

		result = plugin.deregisterNode( mgNoise::id );
		return result;
}
